const puppeteer = require('puppeteer-core')
const stream = require('stream')
const fs = require('fs')
const path = require('path')
const { exit } = require('process')
const { Stream } = require('stream')

const logPrefix = '[videox]'
const deviceModel = puppeteer.devices['iPad']

module.exports = NewInstance

// Exposed videox class constructor.
function NewInstance(options) {
  // Default options.
  this.options = {
    debug: false,
    downloadBrowser: false,
    executePath: '/usr/bin/chromium',
    browserArgs: ['--no-sandbox'],
    logTo: process.stdout,
    downloadAsFile: true,
    downloadPath: '',
    outVideoStream: null,
    outAudioStream: null,
  }

  Object.assign(this.options, options)

  const opts = this.options

  if (process.env.NODE_ENV === 'test') {
    opts.downloadBrowser = true
  }

  this.createdDownloadStream = false

  return this
}

// Open browser and a new page, be prepared for nevigating.
NewInstance.prototype.init = async function () {
  let err

  opts = this.options

  if (opts.downloadBrowser) {
    const browserFetcher = puppeteer.createBrowserFetcher()

    const revisionInfo = await this._to(
      browserFetcher.download('800071'),
      'download browser'
    )

    opts.executePath = revisionInfo.executablePath
    this._log('downloaded browser in path: ' + opts.executePath)
  }

  this.browser = await this._to(
    puppeteer.launch({
      headless: !opts.debug,
      args: opts.browserArgs,
      executablePath: path.resolve(opts.executePath),
    }),
    'launch puppeteer'
  )
}

// Launch download process.
NewInstance.prototype.get = async function (url) {
  const pageUrl = (this.pageUrl = url)

  const opt = this.options

  const page = (this.page = await this._to(
    this.browser.newPage(),
    'open a new page'
  ))

  await this._to(page.emulate(deviceModel), 'emulate device')

  await page.exposeFunction('__videoxLog__', async (str) => {
    this._log(str)
  })

  await page.exposeFunction('__videoxLogRaw__', async (str) => {
    this._logRaw(str)
  })

  await page.exposeFunction('__videoxFatal__', async (str) => {
    await this.destroy()

    throw new Error(str)
  })

  // chunk is base64 string.
  await page.exposeFunction(
    '__videoxWrite__',
    async (mimeCodec, base64Chunk) => {
      const mimeCodecSplited = String.prototype.split.call(mimeCodec, /\/|;/)
      if (mimeCodecSplited.length < 2) {
        throw new Error('invalid type: ' + mimeCodec)
      }

      let mediaType = mimeCodecSplited[0]
      let ext = mimeCodecSplited[1]
      ext = String.prototype.toLowerCase.call(ext)

      if (opt.downloadAsFile && !this.createdDownloadStream) {
        const urlBasename = path.basename(new URL(this.pageUrl).pathname)

        let videoPath = path.resolve(
          opt.downloadPath,
          urlBasename,
          'video' + '.' + ext
        )
        let audioPath = path.resolve(
          opt.downloadPath,
          urlBasename,
          'audio' + '.' + ext
        )

        fs.mkdirSync(path.dirname(videoPath), { recursive: true })

        opt.outVideoStream = fs.createWriteStream(videoPath, { flags: 'w' })
        opt.outAudioStream = fs.createWriteStream(audioPath, { flags: 'w' })

        this.createdDownloadStream = true
      }

      let outStream
      switch (mediaType) {
        case 'video':
          outStream = opt.outVideoStream
          break
        case 'audio':
          outStream = opt.outAudioStream
          break
        default:
          throw new Error('wrong media type: ' + mediaType)
          break
      }

      if (!(outStream instanceof stream.Writable)) {
        throw new Error('outStream not Writable')
      }

      if (!outStream.write(Buffer.from(base64Chunk, 'base64'))) {
        // Wait for it to be drained.
        await new Promise((resolve) => {
          outStream.once('drain', () => resolve())
        })
      }
    }
  )

  // Overwrite some functions MSE would use.
  // See: https://developer.mozilla.org/en-US/docs/Web/API/Media_Source_Extensions_API
  await page.evaluateOnNewDocument(async () => {
    await window.__videoxLog__('evaluateOnNewDocument')

    // Used to record some data.
    window.__videoxObj = {
      mediaSource: null,
      sourceBufferMap: new Map(),
      enumDownloadStatus: {
        DS_UNINTIALIZED: 0,
        DS_DOWNLOADING: 1,
        DS_COMPLETED: 2,
      },
    }

    window.__videoxObj.downloadStatus =
      window.__videoxObj.enumDownloadStatus.DS_UNINTIALIZED

    const addSourceBufferOrigin = MediaSource.prototype.addSourceBuffer
    const endOfStreamOrigin = MediaSource.prototype.endOfStream
    const appendBufferOrigin = SourceBuffer.prototype.appendBuffer

    MediaSource.prototype.addSourceBuffer = function () {
      window.__videoxLog__('addSourceBuffer, mimeCodec: ' + arguments[0])

      const obj = window.__videoxObj

      const sourceBuf = addSourceBufferOrigin.apply(this, arguments)
      sourceBuf.addEventListener('updateend', seek)

      // Record the mediaSource in document.
      if (!obj.mediaSource) {
        obj.mediaSource = this
      } else if (obj.mediaSource !== this) {
        window.__videoxFatal__("mediasource don't match")
      }

      // Record sourceBuffers.
      obj.sourceBufferMap.set(sourceBuf, {
        mimeCodec: arguments[0],
        timeEnd: 0,
      })

      switch (obj.downloadStatus) {
        case obj.enumDownloadStatus.DS_UNINTIALIZED:
          obj.downloadStatus = obj.enumDownloadStatus.DS_DOWNLOADING
          break
        case obj.enumDownloadStatus.DS_DOWNLOADING:
          // There may be multiple sourceBuffer added.
          break
        case obj.enumDownloadStatus.DS_COMPLETED:
        default:
          window.__videoxFatal__(
            'addSourceBuffer wrong download status: ' + obj.downloadStatus
          )
          break
      }

      return sourceBuf
    }

    MediaSource.prototype.endOfStream = function () {
      window.__videoxLog__('endOfStream')

      const obj = window.__videoxObj

      Array.prototype.forEach.call(this.sourceBuffers, (e) => {
        e.removeEventListener('updateend', seek)
      })

      switch (obj.downloadStatus) {
        case obj.enumDownloadStatus.DS_DOWNLOADING:
          obj.downloadStatus = obj.enumDownloadStatus.DS_COMPLETED
          break
        case obj.enumDownloadStatus.DS_COMPLETED:
        case obj.enumDownloadStatus.DS_UNINTIALIZED:
        default:
          window.__videoxFatal__(
            'endOfStream wrong download status: ' + obj.downloadStatus
          )
          break
      }

      return endOfStreamOrigin.apply(this, arguments)
    }

    SourceBuffer.prototype.appendBuffer = function () {
      const obj = window.__videoxObj

      if (obj.downloadStatus !== obj.enumDownloadStatus.DS_DOWNLOADING) {
        window.__videoxFatal__(
          'appendBuffer wrong download status: ' + obj.downloadStatus
        )
      }

      // Get mimeCodec string corresponding to this sourceBuffer instance.
      const sourceBufferObj = obj.sourceBufferMap.get(this)
      if (!sourceBufferObj) {
        window.__videoxFatal__('type not in typeMap')
      }
      if (!sourceBufferObj.mimeCodec) {
        window.__videoxFatal__('empty mimeCodec')
      }

      // Update the buffered time end.
      if (this.buffered.length == 1) {
        const t = this.buffered.end(0)

        sourceBufferObj.timeEnd = t
      }

      // Serialize the bytes chunk. Needed for tranmitting data between browser and program.
      const arr = new Uint8Array(arguments[0])
      const len = arr.byteLength
      let b = ''
      for (let i = 0; i < len; i++) {
        b += String.fromCharCode(arr[i])
      }
      const sed = window.btoa(b)
      window.__videoxWrite__(sourceBufferObj.mimeCodec, sed)

      return appendBufferOrigin.apply(this, arguments)
    }

    function seek() {
      // Place seek operation in next eventloop waiting for updated timeEnd value.
      setTimeout(seekNext, 1)

      // Seek to max buffered time.
      function seekNext() {
        const obj = window.__videoxObj
        if (!obj.mediaSource) {
          window.__videoxFatal__('seek empty mediaSource')
        }

        const sourceBufferMap = obj.sourceBufferMap
        if (sourceBufferMap.size <= 0) {
          window.__videoxFatal__('typemap size 0')
        }

        let timeEndMin = Number.MAX_VALUE
        sourceBufferMap.forEach((val) => {
          if (timeEndMin > val.timeEnd) {
            timeEndMin = val.timeEnd
          }
        })

        const videoEle = document.querySelector('video')
        if (!videoEle) {
          window.__videoxFatal__('videoEle null')
        }

        if (videoEle.currentTime < timeEndMin) {
          videoEle.currentTime = timeEndMin
        }

        // Print download progress in percentage.
        let percent =
          ((videoEle.currentTime / obj.mediaSource.duration) * 100).toFixed(2) +
          '%'
        let str = '\r[videox] download progress: ' + percent
        window.__videoxLogRaw__(str)
      }
    }
  })

  await this._to(
    page.goto(pageUrl, { waitUntil: 'networkidle2' }),
    'page goto: ' + pageUrl
  )

  await this._to(
    page.evaluate(async () => {
      const obj = window.__videoxObj

      const videoEle = document.querySelector('video')
      if (!videoEle) {
        await window.__videoxFatal__('no videoEle')
      }

      if (!String.prototype.startsWith.call(videoEle.src, 'blob')) {
        await window.__videoxFatal__(
          'not MSE, you should download directly from video.src address'
        )
      }

      await window.__videoxLog__('video playing')
      await videoEle.play()

      const checkInterval = 100

      checkLoop: for (;;) {
        switch (obj.downloadStatus) {
          case obj.enumDownloadStatus.DS_UNINTIALIZED:
          case obj.enumDownloadStatus.DS_DOWNLOADING:
            break
          case obj.enumDownloadStatus.DS_COMPLETED:
            break checkLoop
          default:
            window.__videoxFatal__(
              'endOfStream wrong download status: ' + obj.downloadStatus
            )
            break
        }

        await new Promise((resolve) =>
          setTimeout(() => resolve(), checkInterval)
        )
      }
    }),
    'page evaluate'
  )

  await this._to(page.close(), 'close page: ' + pageUrl)
}

// Close the browser.
NewInstance.prototype.destroy = async function () {
  await this._to(this.browser.close(), 'close browser')
}

// Catch promise and log.
NewInstance.prototype._to = async function (pms, msg) {
  let err, data

  this._log(msg + ' ...')
  ;[err, data] = await to(pms)
  if (err) {
    throw new Error(msg + ' error: ' + err)
  }

  this._log(msg + ' OK')

  return data
}

NewInstance.prototype._logRaw = function (str) {
  tlog(this.options.logTo, str, true)
}

NewInstance.prototype._log = function (str) {
  tlog(this.options.logTo, str, false)
}

function tlog(out, msg, raw) {
  if (out && out instanceof stream.Writable) {
    let str

    if (!raw) {
      str = `${logPrefix} ${new Date().toISOString()}: ${msg}\n`
    } else {
      str = msg
    }

    out.write(str)
  }
}

function to(pms) {
  return pms.then((data) => [null, data]).catch((err) => [err, null])
}
